/* wptFileManager.h - File manager routines
 *	Copyright (C) 2001-2005 Timo Schulz
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 2 
 * of the License, or (at your option) any later version.
 *  
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with WinPT; if not, write to the Free Software Foundation, 
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 */

#ifndef WPT_FILE_MANAGER_H
#define WPT_FILE_MANAGER_H

/* File manager commands. */
enum fm_cmd_t {
    FM_NONE	    = 0,
    FM_ENCRYPT	    = 1,	
    FM_SYMENC	    = 2,	
    FM_DECRYPT	    = 3,	
    FM_SIGN	    = 4,
    FM_SIGNENCRYPT  = 5,
    FM_VERIFY	    = 6,
    FM_IMPORT	    = 7,
    FM_LIST	    = 8,
    FM_WIPE	    = 9,
    FM_ENCRYPT_DIR  = 10,
    FM_ENCRYPT_ZIP  = 11
};

/* Sort by constants. */
enum fm_sort_t {
    FM_SORT_NONE    = 0,
    FM_SORT_STAT    = 1,
    FM_SORT_NAME    = 2,
    FM_SORT_OP	    = 3
};

/* File manager handle. */
struct fm_state_s {
    gpgme_ctx_t ctx;		/* currently used gpgme context. */
    gpgme_key_t *recp;		/* selected recipients. */
    int		n_recp;
    char *output;		/* output file name. */
    char *opaque;
    int cancel;			/* 1=if user cancelled operation. */
    HWND dlg;			/* handle to the calling dialog. */
    struct passphrase_cb_s pass_cb;
    struct progress_filter_s *prog_cb;
    int init_cb;
    int cache_cb;
    gpgme_sig_mode_t sigmode;	/* used signature mode. */
    struct {
	unsigned int is_clip:1; /* 1=if clipboard operation. */
	unsigned int revcert:1;
	unsigned int has_seckey:1;
    } import;
    unsigned int wipe:1;	/* 1=if original file should be wiped. */
    unsigned int req_signer:1;	/* 1=if user wants to select a signer. */
};
typedef struct fm_state_s * fm_state_t;


/* Signature verification context for a file. */
struct file_sig_ctx_s {
    char		*file;      /* plaintext file name */
    gpgme_signature_t	sig;	    /* the actual signature */
    unsigned		use_uid:1;  /* 1=if the user id from context should be used. */
    const char		*user_id;
};
typedef struct file_sig_ctx_s *file_sig_ctx_t;


enum file_data_flag_t {
    F_DATA_WRITE    = 0,
    F_DATA_READ	    = 1,    
    F_DATA_NOMAP    = 2
};

/* Gpg file handle. */
struct file_data_s {
    struct gpgme_data_cbs cbs;
    HANDLE	     handle;
    int		     error;
    gpgme_data_t     dat;
    unsigned long    size;
    unsigned long    off;
    void	     *cb_value;
    char	     *name;
};
typedef struct file_data_s *file_data_t;

/*-- wptFileCBS.cpp --*/
gpgme_error_t gpg_file_data_new (const char *fname, int flags, 
				 file_data_t *r_cb);
void	      gpg_file_data_release (file_data_t cb);
void          gpg_file_data_set_cb (file_data_t ctx, 
				    struct progress_filter_s *pfx);

/*-- wptFileManager.cpp --*/
char * fm_quote_file (const char * name);
int fm_parse_command_line (char * cmdl);
int fm_parse_files (listview_ctrl_t lv, HWND dlg, int cmd);
int fm_state_new (fm_state_t *ctx);
void fm_state_release (fm_state_t ctx);
int fm_build (listview_ctrl_t *lv, HWND ctrl);
void fm_delete (listview_ctrl_t lv);
int fm_add_dropped_files (listview_ctrl_t lv, HDROP dd_files);
int fm_add_opened_files (listview_ctrl_t lv, HWND dlg);
void fm_remove_crit_file_attrs (const char *fname, int force);
int fm_get_current_pos (listview_ctrl_t lv);
int fm_sort (listview_ctrl_t lv, int sortby);
void fm_print_md (listview_ctrl_t lv, HWND dlg, int mdalgo);
int fm_assume_onepass_sig (const char * fname);
int fm_check_file_type (listview_ctrl_t lv, int pos, int fm_cmd);
int fm_send_file (listview_ctrl_t lv);

/*-- wptFileVerifyDlg.cpp --*/
void file_verify_add_state (file_sig_ctx_t c);
void file_verify_wait (void);

/*-- ccommands --*/
int fm_encrypt_into_zip (fm_state_t ctx, listview_ctrl_t lv);
int fm_encrypt (fm_state_t c, const char * name, int sign);
int fm_sym_encrypt (fm_state_t c, const char * name);
int fm_decrypt (fm_state_t c, const char * name);
int fm_sign (fm_state_t c, const char * name);
int fm_verify (fm_state_t c, int detached, const char * name);
int fm_import(fm_state_t c, const char *name);
int fm_wipe (const char * name);
int fm_list (const char * name, HWND dlg);
int fm_encrypt_directory (fm_state_t c, const char * name);
int fm_verify_pasted_detsig (listview_ctrl_t lv, HWND dlg);
int fm_assume_onepass_sig (const char * fname);

#endif /* WPT_FILE_MANAGER_H */
